{
  Copyright 2023-2023 Michalis Kamburelis.

  This file is part of "Castle Game Engine".

  "Castle Game Engine" is free software; see the file COPYING.txt,
  included in this distribution, for details about the copyright.

  "Castle Game Engine" is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  ----------------------------------------------------------------------------
}
unit GameViewPlay;

interface

uses Classes,
  CastleComponentSerialize, CastleUIControls, CastleControls, CastleNotifications,
  CastleKeysMouse, CastleViewport, CastleScene, CastleVectors, CastleCameras,
  CastleTransform, GameEnemy, CastleSoundEngine, X3DNodes, X3DFields, CastleRenderOptions;

type
  { Direction expressed a 4 possible horizontal directions.
    In the clockwise order. }
  TDirection = (dirNorth, dirEast, dirSouth, dirWest);

const
  { Convert TDirection to MainViewport.Camera.Direction vector.
    Check these values looking at MainViewport.Camera.Direction (in "All" tab)
    in editor. }
  DirectionVector: array [TDirection] of TVector3 = (
    (X:  0; Y: 0; Z: -1),
    (X:  1; Y: 0; Z:  0),
    (X:  0; Y: 0; Z:  1),
    (X: -1; Y: 0; Z:  0)
  );

  DirectionName: array [TDirection] of String = (
    'North',
    'East',
    'South',
    'West'
  );

  { Time, in seconds, to perform move or rotation. }
  ActionDuration = 0.25;

type
  { Main "playing game" view, where most of the game logic takes place. }
  TViewPlay = class(TCastleView)
  published
    { Components designed using CGE editor.
      These fields will be automatically initialized at Start. }
    MainViewport: TCastleViewport;
    MainFog: TCastleFog;
    SoundShoot: TCastleSound;
    Health: TCastleLabel;
    Ammo: TCastleLabel;
    MapCamera: TCastleCamera;
    EffectTextureField: TSFNode;
    TestTexture1: TImageTextureNode;
    TestTexture2: TImageTextureNode;
    Viewport1: TCastleViewport;

    Monkey: TCastleScene;

    Pistol: TCastleScene;
    AR: TCastleScene;
    Pipe: TCastleScene;
    Shotgun: TCastleScene;

    PortalDetect: TCastleTransform;
    ARDetect: TCastleTransform;
    ShotgunDetect: TCastleTransform;

    AR_to_pick: TCastleScene;
    Shotgun_to_pick: TCastleScene;

    Notifications1: TCastleNotifications;

  private

    { Always synchronized with MainViewport.Camera.Direction. }
    Direction: TDirection;

    { Calculate "Value + Increase"
      doing typecasting as necessary between enums and integers,
      and making sure result is in TDir range. }
    function IncreaseDirection(const Value: TDirection; const Increase: Integer): TDirection;

    procedure Move(const MoveDirection: TDirection);
    procedure Rotate(const RotationChange: Integer);
    procedure ClickMoveLeft(Sender: TObject);
    procedure ClickMoveBackward(Sender: TObject);
    procedure ClickMoveRight(Sender: TObject);
    procedure ClickMoveForward(Sender: TObject);
    procedure ClickRotateLeft(Sender: TObject);
    procedure ClickRotateRight(Sender: TObject);
    procedure SelectWeapon(WeaponNumber: Integer);
    procedure Picked;
  public

    Enemies: TEnemyList;
    constructor Create(AOwner: TComponent); override;
    procedure Start; override;
    procedure Stop; override;
    procedure Update(const SecondsPassed: Single; var HandleInput: Boolean); override;
    function Press(const Event: TInputPressRelease): Boolean; override;
  end;

var
  ViewPlay: TViewPlay;
  Health_P: Integer;
  Ammo_P: Integer;
  Pistol_picked: Boolean;
  AR_picked: Boolean;
  Shotgun_picked: Boolean;

implementation

uses SysUtils, Math,
  CastleLog, CastleStringUtils, CastleFilesUtils, CastleUtils,
  GameViewMenu;

{ TViewPlay ----------------------------------------------------------------- }

constructor TViewPlay.Create(AOwner: TComponent);
begin
  inherited;
  DesignUrl := 'castle-data:/gameviewplay.castle-user-interface';
end;

procedure TViewPlay.Start;
  procedure CreateTextureEffect(const Scene: TCastleScene);
    var
      Effect: TEffectNode;
      EffectPartFragment, EffectPartVertex: TEffectPartNode;
    begin
      Effect := TEffectNode.Create;
      Effect.Language := slGLSL;

      TestTexture1 := TImageTextureNode.Create;
      TestTexture1.SetUrl(['castle-data:/test_textures/handpaintedwall2.png']);
      TestTexture1.KeepExistingBegin;

      TestTexture2 := TImageTextureNode.Create;
      TestTexture2.SetUrl(['castle-data:/test_textures/mountain.png']);
      TestTexture2.KeepExistingBegin;

      EffectTextureField := TSFNode.Create(Effect, true, 'testTexture', [TImageTextureNode], TestTexture1);
      Effect.AddCustomField(EffectTextureField);

      EffectPartFragment := TEffectPartNode.Create;
      EffectPartFragment.ShaderType := stFragment;
      EffectPartFragment.SetUrl(['castle-data:/shaders/texture_effect.fs']);

      EffectPartVertex := TEffectPartNode.Create;
      EffectPartVertex.ShaderType := stVertex;
      EffectPartVertex.SetUrl(['castle-data:/shaders/texture_effect.vs']);

      Effect.SetParts([EffectPartFragment, EffectPartVertex]);
      Scene.RootNode.AddChildren([Effect]);
    end;
var
  SoldierScene: TCastleScene;
  Enemy: TEnemy;
  I: Integer;
begin
  inherited;

  Notifications1 := DesignedComponent('Notifications1') as TCastleNotifications;
  Pistol := DesignedComponent('Pistol') as TCastleScene;
  AR := DesignedComponent('AR') as TCastleScene;
  Shotgun := DesignedComponent('Shotgun') as TCastleScene;
  Pipe := DesignedComponent('Pipe') as TCastleScene;


  AR_to_pick := DesignedComponent('AR_to_pick') as TCastleScene;
  Shotgun_to_pick := DesignedComponent('Shotgun_to_pick') as TCastleScene;
  AR_picked := false;
  Shotgun_picked := false;

  Health_P := 100;
  Ammo_P := 50;
  Enemies := TEnemyList.Create(true);
   for I := 1 to 1 do
   begin
     SoldierScene := DesignedComponent('Enemy' + IntToStr(I)) as TCastleScene;
     { Below using nil as Owner of TEnemy, as the Enemies list already "owns"
       instances of this class, i.e. it will free them. }
     Enemy := TEnemy.Create(nil);
     SoldierScene.AddBehavior(Enemy);
     Enemies.Add(Enemy);
   end;


   Monkey := DesignedComponent('Monkey') as TCastleScene;

   CreateTextureEffect(Monkey);

  Viewport1 := DesignedComponent('Viewport1') as TCastleViewport;

  { Make sure Dir, MainViewport.Camera.Direction and LabelDirection are initially synchronized. }
  Direction := dirNorth;
  MainViewport.Camera.Direction := DirectionVector[Direction];

  Viewport1.Items := MainViewport.Items;
  Viewport1.Camera := MapCamera;

  { At design-time, we keep MainFog.VisibilityRange larger,
    otherwise it makes it hard to actually see level at design-time from top. }
  //MainFog.VisibilityRange := 5;


end;

procedure TViewPlay.Stop;
begin
  FreeAndNil(Enemies);
  TestTexture1.KeepExistingEnd;
  TestTexture2.KeepExistingEnd;
  FreeIfUnusedAndNil(TestTexture1);
  FreeIfUnusedAndNil(TestTexture2);
  inherited;
end;

procedure TViewPlay.Update(const SecondsPassed: Single; var HandleInput: Boolean);
var
  DirectionHorizontal: TVector3;
  Health_L: TCastleLabel;
  Ammo_L: TCastleLabel;
  Pick1: Boolean;
  Pick2: Boolean;
begin
  inherited;

  Health_L := DesignedComponent('Health') as TCastleLabel;
  Ammo_L := DesignedComponent('Ammo') as TCastleLabel;

  Health_L.Caption := IntToStr(Health_P);
  Ammo_L.Caption := IntToStr(Ammo_P);

  Pick1 := true;
  Pick2 := true;

    DirectionHorizontal := MainViewport.Camera.Direction;

    Viewport1.Camera.SetView(
      MainViewport.Camera.WorldTranslation + Vector3(0, 10, 0),
      Vector3(0, -1, 0),
      DirectionHorizontal);

  if PortalDetect.WorldBoundingBox.Contains(MainViewport.Camera.WorldTranslation) then
    begin
      MainViewport.Camera.SetView(Vector3(0.47, 0.51, 1.47), Vector3(-1.0, 0.0, 0.0), Vector3(0.00, 1.00, 0.00));
      Exit;
    end;

    if ARDetect.WorldBoundingBox.Contains(MainViewport.Camera.WorldTranslation) then
    begin
      AR_picked := true;
      AR_to_pick.Exists := false;
      Notifications1.Show('Автомат подобран!');
      Pick1 := false;
      Exit;
    end;

    if ShotgunDetect.WorldBoundingBox.Contains(MainViewport.Camera.WorldTranslation) then
    begin
      Shotgun_picked := true;
      Shotgun_to_pick.Exists := false;
      Notifications1.Show('Дробовик подобран!');
      Pick2 := false;
      Exit;
      end;
end;

function TViewPlay.IncreaseDirection(const Value: TDirection; const Increase: Integer): TDirection;
begin
  Result := TDirection(ChangeIntCycle(Ord(Value), Increase, Ord(High(TDirection))));
end;

procedure TViewPlay.Move(const MoveDirection: TDirection);
const
  GridSize = 1;
var
  NewPos, Pos, Dir, Up: TVector3;
begin
  if MainViewport.Camera.Animation then
    Exit;
  MainViewport.Camera.GetWorldView(Pos, Dir, Up);
  NewPos := Pos + DirectionVector[MoveDirection] * GridSize;
  if not MainViewport.Items.WorldSegmentCollision(Pos, NewPos) then
    MainViewport.Camera.AnimateTo(NewPos, Dir, Up, ActionDuration);
end;

procedure TViewPlay.Rotate(const RotationChange: Integer);
var
  Pos, Dir, Up: TVector3;
begin
  if MainViewport.Camera.Animation then
    Exit;
  MainViewport.Camera.GetWorldView(Pos, Dir, Up);
  Direction := IncreaseDirection(Direction, RotationChange);
  Dir := DirectionVector[Direction];
  MainViewport.Camera.AnimateTo(Pos, Dir, Up, ActionDuration);
end;

function TViewPlay.Press(const Event: TInputPressRelease): Boolean;
var
  HitEnemy: TEnemy;
  Damage: Integer;
begin
  Result := inherited;
  if Result then Exit; // allow the ancestor to handle keys

  if Event.IsMouseButton(buttonLeft) then
  begin
    Pistol.PlayAnimation('shoot', false);
    SoundEngine.Play(SoundShoot);
    Dec(Ammo_P);

    Damage := Random(10);
    Notifications1.Show('Выстрел по врагу');
    Notifications1.Show('Нанесено ' + Damage.ToString + ' урона');

    { We clicked on enemy if
      - TransformUnderMouse indicates we hit something
      - It has a behavior of TEnemy. }
    if (MainViewport.TransformUnderMouse <> nil) and
       (MainViewport.TransformUnderMouse.FindBehavior(TEnemy) <> nil) then
    begin
      HitEnemy := MainViewport.TransformUnderMouse.FindBehavior(TEnemy) as TEnemy;
      HitEnemy.Hurt;
      Notifications1.Show('Враг убит');
    end;

    Exit(true);
  end;

  if not MainViewport.Camera.Animation then
  begin
    if Event.IsKey(keyW) then
    begin
      Move(IncreaseDirection(Direction, 0)); // Move along Direction
      Exit(true);
    end;
    if Event.IsKey(keyS) then
    begin
      Move(IncreaseDirection(Direction, 2)); // Move along the opposite of Direction
      Exit(true);
    end;
    if Event.IsKey(keyD) then
    begin
      Move(IncreaseDirection(Direction, 1)); // Move along Direction rotated to right (i.e. strafe right)
      Exit(true);
    end;
    if Event.IsKey(keyA) then
    begin
      Move(IncreaseDirection(Direction, -1)); // Move along Direction rotated to left (i.e. strafe left)
      Exit(true);
    end;

    if Event.IsKey(keyQ) then
    begin
      Rotate(-1);
      Exit(true);
    end;
    if Event.IsKey(keyE) then
    begin
      Rotate(1);
      Exit(true);
    end;
  end;

  if Event.IsKey(keyF5) then
  begin
    Container.SaveScreenToDefaultFile;
    Exit(true);
  end;

  if Event.IsKey(keyEscape) then
  begin
    Container.View := ViewMenu;
    Exit(true);
  end;

  if Event.IsKey(key1) then
  begin
    SelectWeapon(1);
    Exit(true);
  end;

  if Event.IsKey(key2) then
  begin
    SelectWeapon(2);
    Exit(true);
  end;

  if Event.IsKey(key3) then
  begin
    SelectWeapon(3);
    Exit(true);
  end;

  if Event.IsKey(key4) then
  begin
    SelectWeapon(4);
    Exit(true);
  end;

  if Event.IsKey(keyT) then
  begin
  if EffectTextureField.Value = TestTexture1 then
    EffectTextureField.Send(TestTexture2)
  else
    EffectTextureField.Send(TestTexture1);

  Monkey.GLContextClose;
  end;
end;

procedure TViewPlay.ClickMoveLeft(Sender: TObject);
begin
  Move(IncreaseDirection(Direction, -1));
end;

procedure TViewPlay.ClickMoveBackward(Sender: TObject);
begin
  Move(IncreaseDirection(Direction, 2));
end;

procedure TViewPlay.ClickMoveRight(Sender: TObject);
begin
  Move(IncreaseDirection(Direction, 1));
end;

procedure TViewPlay.ClickMoveForward(Sender: TObject);
begin
  Move(IncreaseDirection(Direction, 0));
end;

procedure TViewPlay.ClickRotateLeft(Sender: TObject);
begin
  Rotate(-1);
end;

procedure TViewPlay.ClickRotateRight(Sender: TObject);
begin
  Rotate(1);
end;

procedure TViewPlay.SelectWeapon(WeaponNumber: Integer);
begin
  if WeaponNumber = 1 then
  begin
    Pipe.Exists := true;
    Pistol.Exists := false;
    AR.Exists := false;
    Shotgun.Exists := false;
  end;

  if WeaponNumber = 2 then
  begin
    Pipe.Exists := false;
    Pistol.Exists := true;
    AR.Exists := false;
    Shotgun.Exists := false;
  end;

  if WeaponNumber = 3 then
  begin
  if Shotgun_picked = true then
  begin
    Pipe.Exists := false;
    Pistol.Exists := false;
    AR.Exists := false;
    Shotgun.Exists := true;
    end;
  end;

  if WeaponNumber = 4 then
  begin
  if AR_picked = true then
  begin
    Pipe.Exists := false;
    Pistol.Exists := false;
    AR.Exists := true;
    Shotgun.Exists := false;
  end;
  end;
end;

end.
